"""test Face class."""
from honeybee.face import Face
from honeybee.properties import face_types as Types
from honeybee.boundarycondition import boundary_conditions
import pytest


def test_name():
    vertices = [[0, 0, 0], [0, 10, 0], [0, 10, 3], [0, 0, 3]]
    f = Face.from_vertices('Wall: SIM_INT_AIR [920980]', vertices)
    assert f.name == 'WallSIM_INT_AIR920980'
    assert f.name_original == 'Wall: SIM_INT_AIR [920980]'


def test_vertices_count():
    with pytest.raises(Exception):
        Face.from_vertices('no vertices', [], Types.wall)


def test_type_from_vertices():
    vertices_wall = [[0, 0, 0], [0, 10, 0], [0, 10, 3], [0, 0, 3]]
    vertices_floor = [[0, 0, 0], [0, 10, 0], [10, 10, 0], [10, 0, 0]]
    vertices_roof = list(reversed(vertices_floor))

    wf = Face.from_vertices('wall', vertices_wall)
    assert wf.properties.face_type == wf.properties.TYPES.wall

    rf = Face.from_vertices('roof', vertices_roof)
    assert rf.properties.face_type == rf.properties.TYPES.roof_ceiling

    ff = Face.from_vertices('floor', vertices_floor)
    assert ff.properties.face_type == ff.properties.TYPES.floor


def test_default_boundary_condition():
    vertices = [[0, 0, 0], [0, 10, 0], [0, 10, 3], [0, 0, 3]]
    face = Face.from_vertices('wall', vertices)
    assert face.boundary_condition == boundary_conditions.outdoors


def test_set_boundary_condition():
    vertices = [[0, 0, 0], [0, 10, 0], [0, 10, 3], [0, 0, 3]]
    face = Face.from_vertices('wall', vertices, Types.wall, boundary_conditions.ground)
    assert face.boundary_condition == boundary_conditions.ground

    face.boundary_condition = boundary_conditions.outdoors
    assert face.boundary_condition == boundary_conditions.outdoors


def test_to_dict():
    vertices = [[0, 0, 0], [0, 10, 0], [0, 10, 3], [0, 0, 3]]
    face = Face.from_vertices('wall', vertices, Types.wall, boundary_conditions.ground)
    
    fd = face.to_dict([])
    assert fd['type'] == 'Face'
    assert fd['name'] == 'wall'
    assert 'vertices' in fd
    assert len(fd['vertices']) == len(vertices)
    assert 'properties' in fd
    assert fd['properties']['type'] == 'FaceProperties'
    assert fd['properties']['face_type'] == 'Wall'
    assert fd['properties']['boundary_condition']['name'] == 'Ground'
    assert fd['properties']['boundary_condition']['bc_object'] == ''



"""Face Properties."""
from .facetype import face_types
from .boundarycondition import _BoundaryCondition, boundary_conditions


class Properties(object):
    """Geometry Properties.

    Base class for geometry properties. This class will be extended by plugins.

    prop = Properties(srf_type, boundary_condition)
    prop.radiance -> RadianceProperties
    prop.energy -> EnergyProperties
    """
    TYPES = face_types

    def __init__(self, face_type, boundary_condition=None):
        self.face_type = face_type
        self.boundary_condition = boundary_condition or boundary_conditions.outdoors

    @property
    def face_type(self):
        """Get and set Surface Type."""
        return self._face_type

    @face_type.setter
    def face_type(self, value):
        assert value in self.TYPES, '{} is not a valid type.'.format(value)
        self._face_type = value

    @property
    def boundary_condition(self):
        """Get and set boundary condition."""
        return self._boundary_condition

    @boundary_condition.setter
    def boundary_condition(self, bc):
        assert isinstance(bc, _BoundaryCondition), \
            'Expected BoundaryCondition not {}'.format(type(bc))
        self._boundary_condition = bc

    def to_dict(self, include=None):
        """convert properties to dictionary.

        args:
            include: A list of keys to be included in dictionary besides face_type
                and boundary condition. If None all the available keys will be included.
        """
        base = {
            'type': 'FaceProperties',
            'face_type': self.face_type.name
        }

        if include is not None:
            attr = include
        else:
            attr = [atr for atr in dir(self) if not atr.startswith('_')]

        for atr in attr:
            var = getattr(self, atr)
            if not hasattr(var, 'to_dict'):
                continue
            try:
                base.update(var.to_dict())
            except Exception as e:
                raise Exception(
                    'Failed to convert {} to a dict: {}'.format(var, e)
                )

        if 'energy' in base:
            base['boundary_condition'] = self.boundary_condition.to_dict(full=True)
        else:
            base['boundary_condition'] = self.boundary_condition.to_dict(full=False)
        return base

    def ToString(self):
        """Overwrite .NET ToString method."""
        return self.__repr__()

    def __repr__(self):
        """Properties representation."""
        return 'FaceProperties:%s' % str(self.face_type.name)



"""Energy Types."""
from ladybug_geometry.geometry3d.pointvector import Vector3D
import math


class _FaceType(object):
    @property
    def name(self):
        return self.__class__.__name__

    def __repr__(self):
        return self.name


class Wall(_FaceType):
    pass


class RoofCeiling(_FaceType):
    pass


class Floor(_FaceType):
    pass


class AirWall(_FaceType):
    pass


class Shading(_FaceType):
    pass


class _Types(object):
    """Face types."""

    def __init__(self):
        self._wall = Wall()
        self._roof_ceiling = RoofCeiling()
        self._floor = Floor()
        self._air_wall = AirWall()
        self._shading = Shading()

    @property
    def wall(self):
        return self._wall

    @property
    def roof_ceiling(self):
        return self._roof_ceiling

    @property
    def floor(self):
        return self._floor

    @property
    def airwall(self):
        return self._air_wall

    @property
    def shading(self):
        return self._shading

    def __contains__(self, value):
        return isinstance(value, _FaceType)


face_types = _Types()


def get_type_from_normal(normal_vector, roof_angle=30, floor_angle=150):
    """Return face type based on the angle between Z axis and normal vector.

    Angles between 0 and roof_angle will be set to roof_ceiling.
    Angles between roof_angle and floor_angle will be set to wall.
    Angles larger than floor angle will be set to floor.

    args:
        normal_vector: Normal vector as a ladybug Vector3D.
        roof_angle: Cutting angle for roof from Z axis (default: 30).
        floor_angle: Cutting angle for floor from Z axis (default: 150).
    
    returns:
        face type.
    """

    roof_angle = math.radians(roof_angle)
    floor_angle = math.radians(floor_angle)
    z_axis = Vector3D(0, 0, 1)
    angle = z_axis.angle(normal_vector)
    if angle < roof_angle:
        return face_types.roof_ceiling
    elif roof_angle <= angle < floor_angle:
        return face_types.wall
    else:
        return face_types.floor

    return face_types.wall



"""Face Boundary Condition."""
import weakref


class _BoundaryCondition(object):

    __slots__ = (
        '_sun_exposure', '_wind_exposure', '_boundary_condition_object', '_view_factor'
    )

    def __init__(self):
        self.sun_exposure = False
        self.wind_exposure = False
        self.boundary_condition_object = None
        self.view_factor = 'autocalculate'

    @property
    def name(self):
        return self.__class__.__name__

    @property
    def sun_exposure(self):
        return self._sun_exposure

    @sun_exposure.setter
    def sun_exposure(self, value):
        assert isinstance(value, bool), \
            'Input value must be a Boolean not {}.'.format(type(value))
        self._sun_exposure = value

    @property
    def wind_exposure(self):
        return self._wind_exposure

    @wind_exposure.setter
    def wind_exposure(self, value):
        assert isinstance(value, bool), \
            'Input value must be a Boolean not {}.'.format(type(value))
        self._wind_exposure = value

    @property
    def boundary_condition_object(self):
        return self._boundary_condition_object

    @boundary_condition_object.setter
    def boundary_condition_object(self, bco):
        self._boundary_condition_object = bco

    @property
    def view_factor(self):
        return self._view_factor

    @view_factor.setter
    def view_factor(self, vf):
        if vf == 'autocalculate':
            self._view_factor = vf
        else:
            self._view_factor = float(vf)

    @property
    def sun_exposure_idf(self):
        return 'NoSun' if not self.wind_exposure else 'SunExposed'

    @property
    def wind_exposure_idf(self):
        return 'NoWind' if not self.wind_exposure else 'WindExposed'

    @property
    def boundary_condition_object_idf(self):
        return '' if not self.boundary_condition_object else \
            self.boundary_condition_object.name

    def to_dict(self, full=False):
        """Boundary condition as a dictionary.
        
        args:
            full: Set to True to get the full dictionary which includes energy simulation
            specific keys such as sun_exposure, wind_exposure and view_factor. Default is
            set to False.
        
        returns:
            A dictionary.
        """
        if full:
            return {
                'name': self.name,
                'bc_object': self.boundary_condition_object_idf,
                'sun_exposure': self.sun_exposure_idf,
                'wind_exposure': self.wind_exposure_idf,
                'view_factor': self.view_factor
            }
        else:
            return {
                'name': self.name,
                'bc_object': self.boundary_condition_object_idf
            }

    def __repr__(self):
        return self.name


class Surface(_BoundaryCondition):

    __slots__ = ()

    def __init__(self, other_object=None):
        _BoundaryCondition.__init__(self)
        if other_object:
            self.boundary_condition_object = weakref.proxy(other_object)


class Outdoors(_BoundaryCondition):

    __slots__ = ()

    def __init__(self):
        _BoundaryCondition.__init__(self)
        self.wind_exposure = True
        self.sun_exposure = True


class Ground(_BoundaryCondition):
    __slots__ = ()
    pass


class _BoundaryConditions(object):
    """Boundary conditions."""

    def __init__(self):
        self._outdoors = Outdoors()
        self._ground = Ground()

    @property
    def outdoors(self):
        return self._outdoors

    @property
    def ground(self):
        return self._ground

    def surface(self, other_object=None):
        return Surface(other_object)

    def __contains__(self, value):
        return isinstance(value, _BoundaryCondition)


boundary_conditions = _BoundaryConditions()



"""test Face class."""
from honeybee.boundarycondition import boundary_conditions
from honeybee.face import Face
import pytest

bcs = boundary_conditions


def test_outdoors():
    bc = bcs.outdoors
    assert bc.name == 'Outdoors'
    assert bc.sun_exposure == True
    assert bc.sun_exposure_idf == 'SunExposed'
    assert bc.wind_exposure == True
    assert bc.wind_exposure_idf == 'WindExposed'
    assert bc.boundary_condition_object == None
    assert bc.boundary_condition_object_idf == ''


def test_outdoors_to_dict():
    bc = bcs.outdoors
    outdict = bc.to_dict(full=True)
    assert outdict['name'] == 'Outdoors'
    assert outdict['bc_object'] == ''
    assert outdict['sun_exposure'] == 'SunExposed'
    assert outdict['wind_exposure'] == 'WindExposed'
    assert outdict['view_factor'] == 'autocalculate'
    outdict = bc.to_dict(full=False)
    assert 'sun_exposure' not in outdict
    assert 'wind_exposure' not in outdict
    assert 'view_factor' not in outdict


def test_ground():
    bc = bcs.ground
    assert bc.name == 'Ground'
    assert bc.sun_exposure == False
    assert bc.sun_exposure_idf == 'NoSun'
    assert bc.wind_exposure == False
    assert bc.wind_exposure_idf == 'NoWind'
    assert bc.boundary_condition_object == None
    assert bc.boundary_condition_object_idf == ''


def test_surface():
    vertices = [[0, 0, 0], [0, 10, 0], [0, 10, 3], [0, 0, 3]]
    face = Face.from_vertices('wall', vertices)
    bc = bcs.surface(face)
    assert bc.name == 'Surface'
    assert bc.sun_exposure == False
    assert bc.sun_exposure_idf == 'NoSun'
    assert bc.wind_exposure == False
    assert bc.wind_exposure_idf == 'NoWind'
    assert bc.boundary_condition_object.name == face.name
    assert bc.boundary_condition_object_idf == face.name



# coding: utf-8
"""Honeybee Face."""
from .properties import Properties
from .facetype import get_type_from_normal
import honeybee.writer as writer
from ladybug_geometry.geometry3d.face import Face3D
from ladybug_geometry.geometry3d.pointvector import Point3D

import re
import weakref


class Face(object):
    """A single planar face."""

    def __init__(self, name, geometry, face_type=None, boundary_condition=None):
        """A single planar face.
        Args:
            name: Face name.
            geometry: A ladybug-geometry Face3D.
            face_type: Face type (e.g. wall, floor).
            boundary_condition: Face boundary condition (Outdoors, Ground, etc.)
        """
        self.name = name
        assert isinstance(geometry, Face3D), \
            'Expected ladybug_geometry Face3D not {}'.format(type(geometry))
        self._geometry = geometry
        # _parent will be set when the Face is added to a Zone
        # in case of aperture it will be added when aperture is added to a Face.
        self._parent = None
        # get face type based on normal if face_type is not provided
        face_type = face_type or get_type_from_normal(geometry.normal)
        self._properties = Properties(face_type, boundary_condition)
        self._writer = writer
        self._apertures = []

    @property
    def name(self):
        """Face name."""
        return self._name

    @name.setter
    def name(self, value):
        self._name = re.sub(r'[^.A-Za-z0-9_-]', '', value)
        self._name_original = value

    @property
    def name_original(self):
        """Original input name by user.

        If there is no illegal characters in name then name and name_original will be the
        same. Legal characters are ., A-Z, a-z, 0-9, _ and -. Invalid characters are
        removed from the original name for compatability with simulation engines.
        """
        return self._name_original

    @property
    def vertices(self):
        """List of vertices."""
        return self._geometry.vertices

    @property
    def face_type(self):
        """Face type."""
        return self._properties.face_type

    @face_type.setter
    def face_type(self, f_type):
        self._properties.face_type = f_type

    @property
    def boundary_condition(self):
        """Face type."""
        return self._properties.boundary_condition

    @boundary_condition.setter
    def boundary_condition(self, bc):
        self._properties.boundary_condition = bc
    
    @property
    def apertures(self):
        """List of apertures."""
        return self._apertures

    @property
    def parent(self):
        """Parent zone."""
        return self._parent

    @classmethod
    def from_vertices(cls, name, vertices, face_type=None, boundary_condition=None):
        """Create a Face from vertices.

        args:
            name: Face name.
            vertices: A flattened list of 3 or more vertices as (x, y, z).
            face_type: Face type (e.g. wall, floor).
        """
        geometry = Face3D([Point3D(*v) for v in vertices])
        return cls(name, geometry, face_type, boundary_condition)

    @property
    def properties(self):
        """Face properties.

        Radiance, energy and other face properties.
        """
        return self._properties

    @property
    def to(self):
        """Face writer object.

        Use this method to access Writer class to write the face in different formats.

        face.to.idf(face) -> idf string.
        face.to.radiance(face) -> Radiance string.
        """
        return self._writer

    def to_dict(self, included_prop=None):
        """Return Face as a dictionary.
        
        args:
            included_prop: Use properties filter to filter keys that must be included in
            output dictionary. For example ['energy'] will include 'energy' key if
            available in properties to_dict. By default all the keys will be included.
            To exclude all the keys from plugins use an empty list.
        """
        base = {
            'type': self.__class__.__name__,
            'name': self.name,
            'name_original': self.name_original,
            'vertices': [{'x': ver.x, 'y': ver.y, 'z': ver.z} for ver in self.vertices],
            'properties': self.properties.to_dict(included_prop)
        }

        if self.parent:
            base['parent'] = {'name': self.parent.name}

        if self.apertures:
            base['apertures'] = [ap.to_dict(included_prop) for ap in self.apertures]

        return base


    def add_aperture(self, aperture):
        """Add an aperture to face."""
        aperture._parent = self
        self._apertures.append(weakref.proxy(aperture))

    def ToString(self):
        return self.__repr__()

    def __repr__(self):
        return 'Face:%s' % self.name_original
